#include<ctype.h>
#include<sys/stat.h>
#include<sys/types.h>
#include<stdio.h>
#include<utime.h>
#include<unistd.h>

#include"Configure.h"
#include"Logger.h"
#include"Error.h"
#include"NVNewsgroup.h"

#ifdef ASSERT_ON
void NVNewsgroup::testdb(void)
{
  //  VERB(log.p(Logger::Debug) << "NVNewsgroup::testdb()\n"); 
  char *data;
  size_t szdata;
  unsigned long f,i,l;
  Article *art=NULL;
  int corr=0;
  
  NVArray::lock(NVcontainer::ShrdLock);
  NVArray::getsize(&f,&l);
  for(i=f;i<=l;i++) {
    NVArray::sget(i,&data,&szdata);
    if(data && *data!=overview && *data!=article && *data!=bigarticle) {
      corr++;
    }
  }
  if(corr) {
    VERB(log.p(Logger::Notice) << "NVNewsgroup::testdb: Oops, "<<corr<<" errors found in database!\n");
  }
  NVArray::lock(NVcontainer::UnLock);
}
#endif

void NVNewsgroup::sprintover(ostream &os,unsigned int nbr)
{
  char *data;
  size_t szdata;
  unsigned long f,l;
  String over;

  NVArray::getsize(&f,&l);
  if(f<=nbr && nbr<=l) {
    NVArray::sget(nbr,&data,&szdata);
    if(data) {
      if(*data==article) {
	Article *art;
	//FIX! This is highly inefficient!
	//FIX! The article is copied and all the overview data is copied 
	//FIX! again. We should allow Articles to use externally provided 
	//FIX! memory (eg. memory from the NVNewsgroup).
	// szdata ... Size of article including the trailing \0
	art=new Article(nbr,data+sizeof(char),szdata-sizeof(char));
	_OverviewFormat->convert(art,over);
	os << over << "\r\n";
      } else if(*data==bigarticle || *data==overview) {
	os << data+sizeof(char) << "\r\n";
      } else {
	// Impossible
      }
    }
  }
}

void NVNewsgroup::getsize(unsigned int *f, unsigned int *l) 
{
  *f=arrfst; *l=arrlst;
}

void NVNewsgroup::setsize(unsigned int f, unsigned int l)
{
  VERB(log.p(Logger::Debug) << "NVNewsgroup::setsize("<<f<<","<<l<<")\n"); 
  char fn[MAXPATHLEN];
  unsigned int i,n;

  lock(NVcontainer::ExclLock);  
  if(arrfst<=arrlst) {
    if(f==arrfst && l==arrlst) return;
    lock(NVcontainer::ExclLock);  
    // free all those elements from arrfst -> f
    n=f-arrfst;
    for(i=0;i<n;i++) {
      if(arrtab[i]) {
	if(*(mem_p+arrtab[i]+sizeof(long))==bigarticle) {
	  // Remove article from disk
	  sprintf(fn,"%s/.art%ud",_SpoolDirectory,i);
	  unlink(fn);
	}
	nvfree(arrtab[i]);
      }
    }
    
    // free all those elements from l -> arrlst
    n=arrlst-arrfst+1;
    for(i=l-arrfst+1;i<n;i++) {
      if(arrtab[i]) {
	if(*(mem_p+arrtab[i]+sizeof(long))==bigarticle) {
	  // Remove article from disk
	  sprintf(fn,"%s/.art%ud",_SpoolDirectory,i);
	  unlink(fn);
	}
	nvfree(arrtab[i]);
      }
    }
  }
  // resize array
  NVArray::ssetsize(f,l);
  lock(NVcontainer::UnLock);  
}

Article *NVNewsgroup::getarticle(unsigned int nbr)
{
  //  VERB(log.p(Logger::Debug) << "NVNewsgroup::getarticle(over)\n"); 

  char *data;
  size_t szdata;
  unsigned long f,l;
  Article *art=NULL;

  NVArray::lock(NVcontainer::ShrdLock);
  NVArray::getsize(&f,&l);
  if(nbr>=f && nbr<=l) {
    NVArray::sget(nbr,&data,&szdata);
    if(data) {
      if(*data==article) {
	// szdata ... Size of article including the trailing \0
	art=new Article(nbr,data+sizeof(char),szdata-sizeof(char));
      } else if(*data==bigarticle) {
	char fn[MAXPATHLEN];
	ifstream fs;
	sprintf(fn,"%s/.art%u",_SpoolDirectory,nbr);
	fs.open(fn);
	art=new Article(nbr);
	art->read(fs);
	fs.close();
      } else {
	// Article not available
	if((art=retrievearticle(nbr))!=NULL) setarticle(art);
      }
    } else {
      // Article not available
      if((art=retrievearticle(nbr))!=NULL) setarticle(art);
    }
  }
  NVArray::lock(NVcontainer::UnLock);
  return art;
}

void NVNewsgroup::setarticle(Article *art)
{
  unsigned long f,l;
  unsigned int nbr;

  char buf[1024];
  sprintf(buf,"NVNewsgroup::setarticle(*art(nbr=%d))\n",art->getnbr());
  VERB(log.p(Logger::Notice) << buf); 
  ASSERT(testdb());

  nbr=art->getnbr();
  NVArray::lock(NVcontainer::ExclLock);
  NVArray::getsize(&f,&l);
  if(f<=nbr && nbr<=l) {
    char *data;
    size_t szdata;
    unsigned long i;

    i=nbr-arrfst;
    NVArray::sget(nbr,&data,&szdata);
    if(art->length()<=15360 && (!data || *data==overview)) {
      // Article smaller/equal 15k => Store article in NewsgroupDB
      nvoff_t x;
      if(data) {
	nvfree(arrtab[i]);
      }
      szdata=art->length()+1+sizeof(char);
      x=nvalloc(szdata+sizeof(unsigned long));
      arrtab[i]=x;
      *((unsigned long*)(mem_p+x))=szdata;
      data=mem_p+x+sizeof(unsigned long);
      *data=article;
      sprintf(buf,"NVN...cle: Before memcpy arrtab=%p data=%p\n",arrtab,data);
      VERB(log.p(Logger::Notice) << buf);
      ASSERT(testdb());
      memcpy(data+sizeof(char),art->gettext(),art->length()+1);
    } else if(art->length()>15360 && (!data || *data==overview)) {
      // Article bigger 15k => Store externally
      if(!data) {
	String over;
	nvoff_t x;
	_OverviewFormat->convert(art,over);
	szdata=over.length()+1+sizeof(char);
	x=nvalloc(szdata+sizeof(unsigned long));
	arrtab[i]=x;
	*((unsigned long*)(mem_p+x))=szdata;
	data=mem_p+x+sizeof(unsigned long);
	memcpy(data+sizeof(char),(const char*)over,over.length()+1);
      }
      *data=bigarticle;
      char fn[MAXPATHLEN];
      ofstream fs;
      sprintf(fn,"%s/.art%u",_SpoolDirectory,nbr);
      fs.open(fn);
      fs << *art;
      fs.close();
      *data=bigarticle;
    } else {
      VERB(log.p(Logger::Debug) << "NVNewsgroup::setarticle: Tried to store an existing article\n"); 
    }
  } /* if(f<=nbr<=l) */
  lock(NVcontainer::UnLock);
  ASSERT(testdb());
}

void NVNewsgroup::printarticle(ostream &os,unsigned int nbr)
{
  char *data;
  size_t szdata;
  unsigned long f,l;

  NVArray::lock(NVcontainer::ShrdLock);
  NVArray::getsize(&f,&l);
  if(nbr>=f && nbr<=l) {
    NVArray::sget(nbr,&data,&szdata);
    if(data) {
      if(*data==article) {
	os << data+sizeof(char);
      } else if(*data==bigarticle) {
	Article art(nbr);
	char fn[MAXPATHLEN];
	ifstream fs;
	sprintf(fn,"%s/.art%u",_SpoolDirectory,nbr);
	fs.open(fn);
	art.read(fs);
	fs.close();
	os << art;
      } else {
	// Article not available
	Article *art;
	if((art=retrievearticle(nbr))!=NULL) setarticle(art);
	os << *art;
      }
    }
  }
  NVArray::lock(NVcontainer::UnLock);
}

String NVNewsgroup::getover(unsigned int nbr) {
  //  VERB(log.p(Logger::Debug) << "NVNewsgroup::getover(over)\n"); 

  char *data;
  size_t szdata;
  unsigned long f,l;
  String over;

  NVArray::lock(NVcontainer::ShrdLock);
  NVArray::getsize(&f,&l);
  if(f<=nbr && nbr<=l) {
    NVArray::sget(nbr,&data,&szdata);
    if(data) {
      if(*data==article) {
	Article *art;
	//FIX! This is inefficient!
	//FIX! The article is copied and all the overview data is copied 
	//FIX! again. We should allow Articles to use externally provided 
	//FIX! memory (eg. memory from the NVNewsgroup).
	// szdata ... Size of article including the trailing \0 plus
	//            1char (= InformationTag of NVNewsgroup)
	art=new Article(nbr,data+sizeof(char),szdata-sizeof(char));
	_OverviewFormat->convert(art,over);
      } else if(*data==bigarticle || *data==overview) {
	over=data+sizeof(char);
      } else {
	// Impossible
	VERB(log.p(Logger::Critical) <<"NVNewsgroup::getover. Oops, reached ?dead? code 936!!!\n");
      }
    }
  }
  NVArray::lock(NVcontainer::UnLock);
  return over;
}

void NVNewsgroup::setover(String over)
{
  //  VERB(log.p(Logger::Notice) << "NVNewsgroup::setover(over)\n"); 

  unsigned long f,l,nbr;
  
  nbr=atoi((const char*)over);
  NVArray::lock(NVcontainer::ExclLock);
  NVArray::getsize(&f,&l);
  if(f<=nbr && nbr<=l) {
    char *data;
    size_t szdata;
    unsigned long i;

    i=nbr-arrfst;
    NVArray::sget(nbr,&data,&szdata);
    if(!data) {
      szdata=over.length()+1+sizeof(char);
      nvoff_t x=nvalloc(szdata+sizeof(unsigned long));
      arrtab[i]=x;
      *((unsigned long*)(mem_p+x))=szdata;
      data=mem_p+x+sizeof(unsigned long);
      memcpy(data+sizeof(char),(const char*)over,over.length()+1);
      *data=overview;
    }
  }
  NVArray::lock(NVcontainer::UnLock);
  ASSERT(testdb());
}

void NVNewsgroup::readoverdb(istream &is)
{
  TRACE(ctrace << "NVNewsgroup::readoverdb(&is)\n");
  Regex r("[\r\n]+");
  String line1,line2;
  time_t now;
  int i;

  lock(NVcontainer::ExclLock);
  for(;;) {
    readline(is,line1,'\n',0);
    if(line1==".\r\n" || line1==".\n") break;
    if(is.eof()) break;
    line1.at(r,-1)="";
    if(_OverviewFormat->dotrans) {
      // Convert Record
      _OverviewFormat->convert(line1,line2);
      setover(line2);
    } else {
      setover(line1);
    }
  }

  time(&now);
  setmtime(now);
  lock(NVcontainer::UnLock);
}

void NVNewsgroup::printoverdb(ostream &os, unsigned int f, unsigned int l)
{
  unsigned int i;
  lock(NVcontainer::ShrdLock);
  if(f<arrfst) f=arrfst;
  if(arrlst<l) l=arrlst;
  for(i=f;i<=l;i++) sprintover(os,i);
  lock(NVcontainer::UnLock);
}

void NVNewsgroup::printheaderdb(ostream &os, 
				const char *header, 
				unsigned int f=0, unsigned int l=UINT_MAX)
{
  char *data,xheader[512],*p;
  const char *fld,*q;
  size_t szdata;
  unsigned long gf,i,gl;

  p=xheader; q=header;
  while((*p++=*q++));
  *(p-1)=':';
  *p='\0';
  NVArray::lock(NVcontainer::ShrdLock);
  NVArray::getsize(&gf,&gl);
  if(f<gf) f=gf;
  if(l>gl) l=gl;
  for(i=f;i<=l;i++) {
    NVArray::sget(i,&data,&szdata);
    if(data) {
      if(*data==article) {
	Article *art;
	//FIX! This is inefficient!
	//FIX! The article is copied and all the overview data is copied 
	//FIX! again. We should allow Articles to use externally provided 
	//FIX! memory (eg. memory from the NVNewsgroup).
	// szdata ... Size of article including the trailing \0 plus
	//            1char (= InformationTag of NVNewsgroup)
	art=new Article(i,data+sizeof(char),szdata-sizeof(char));
	if((fld=art->getfield(xheader))==NULL) {
	  os << i << " (none)\r\n";
	} else {
	  os << i << " " << fld << "\r\n";
	}
      } else if(*data==bigarticle || *data==overview) {
	const char *over=data+sizeof(char);
	if((fld=_OverviewFormat->getfield(over,xheader))==NULL) {
	  os << i << " (none)\r\n";
	} else {
	  os << i << " " << fld << "\r\n";
	}
      }
    }
  }
  NVArray::lock(NVcontainer::UnLock);
}

void NVNewsgroup::printlistgroup(ostream &os)
{
  TRACE(ctrace << "NVNewsgroup::printlistgroup(&os)\n");
  unsigned int i,f,l;
  lock(NVcontainer::ShrdLock);
  for(i=arrfst;i<=arrlst;i++) {
    if(shas_element(i)) os << i << endl;
  }
  lock(NVcontainer::UnLock);  
}
